import logging
import os

import numpy as np
import pandas as pd
import tensorflow as tf
from google.cloud import storage


def get_bucket():
    """Get information on Google cloud bucket"""
    storage_client = storage.Client()
    buckets = list(storage_client.list_buckets())
    bucket_name = buckets[0].name
    bucket = storage_client.get_bucket(bucket_name)
    return bucket


def read_and_expand(filename, path, labels):
    """read in data and expand data set by label placeholders"""
    print("Loading data from " + path + "/" + filename + ".csv ...")
    input_file = pd.read_csv(path + "/" + filename + ".csv", header=None)  # neither index nor header in csv file
    #print("Data read successfully")
    logging.info('Data read successfully')
    if labels:
        print("with labels")
        points = input_file.iloc[:, :-1]
        labels = input_file.iloc[:, -1]
        dataset = tf.convert_to_tensor(points, dtype=tf.float32)
        true_labels = tf.convert_to_tensor(labels, dtype=tf.float32)
    else:
        dataset = tf.convert_to_tensor(input_file, dtype=tf.float32)
        true_labels = None
    n_objects, n_features = dataset.shape
    placeholder_labels = tf.zeros(shape=(n_objects, 1), dtype=tf.float32)
    data_expanded = tf.concat([dataset, placeholder_labels], 1)
    logging.info('Data expanded by label placeholder')
    return data_expanded, true_labels, n_objects, n_features


def batch_dataset(dataset, batch_size_per_replica, n_objects, num_replicas):
    """ return data sets of per-replica batch size and remainder if exists"""
    global_batch_size = batch_size_per_replica * num_replicas
    dataset_tf = tf.data.Dataset.from_tensor_slices(dataset)
    # dataset_shuffled = dataset_tf.shuffle(buffer_size=num_samples*2)
    # dataset_batched = dataset_shuffled.batch(global_batch_size, drop_remainder=True)

    dataset_batched = dataset_tf.batch(global_batch_size, drop_remainder=True)
    logging.info('Dataset is batched with a global batch size of {} and {} batches; batch size per replica = {}'
                 .format(global_batch_size, n_objects // global_batch_size, batch_size_per_replica))
    if n_objects % global_batch_size != 0:
        num_elems = n_objects % global_batch_size
        print("Remainder of {} elements exists".format(num_elems))
        remainder = dataset[-num_elems:]  # or dataset_shuffled
    else:
        remainder = None
    return dataset_batched, remainder


def save_results(data, centroids_final, timestamp, input_filename, directory):
    """ save clustering results"""
    print("Saving results ...")
    # save centroids
    time_formatted = timestamp.strftime("%Y-%m-%d_%H-%M-%S")
    filename_centroids = 'final_centroids_' + input_filename + '_' + time_formatted + '.csv'
    np.savetxt(directory + '/' + filename_centroids, centroids_final, delimiter=',')


    # save data with labels
    filename_final = 'data_final_labels_' + input_filename + '_' + time_formatted + '.csv'
    np.savetxt(directory + '/' + filename_final, data, delimiter=',')
    print("Results saved")
    return filename_centroids, filename_final


def upload_results(directory, filename_initial_centroids, filename_final_centroids, filename_final):
    """upload clustering results to Google cloud bucket"""
    print("Uploading results ...")
    bucket = get_bucket()

    f = os.path.join(directory, filename_initial_centroids)
    if os.path.isfile(f):
        blob = bucket.blob("processed/" + filename_initial_centroids)
        blob.content_type = "text/csv"
        blob.upload_from_filename(f)

    f = os.path.join(directory, filename_final_centroids)
    if os.path.isfile(f):
        blob = bucket.blob("processed/" + filename_final_centroids)
        blob.content_type = "text/csv"
        blob.upload_from_filename(f)

    f = os.path.join(directory, filename_final)
    if os.path.isfile(f):
        blob = bucket.blob("processed/" + filename_final)
        blob.content_type = "text/csv"
        blob.upload_from_filename(f)

    print("Results uploaded to Google Cloud Bucket")
    return
