import csv
import logging
import os.path

import numpy as np
import tensorflow as tf

from google.cloud import storage
from sklearn.metrics import adjusted_mutual_info_score, adjusted_rand_score, calinski_harabasz_score, \
    davies_bouldin_score


def compute_sse(input_data, centroids, n_clusters):
    """return sse"""
    points = input_data[:, :-1]
    labels = input_data[:, -1]
    distance = np.zeros(points.shape[0])
    for k in range(n_clusters):
        distance[labels == k] = np.linalg.norm(points[labels == k] - centroids[k], axis=1)
    return np.sum(np.square(distance)) / points.shape[0]


def compute_ch(data):
    """return calinski_harabasz score"""
    return calinski_harabasz_score(data[:, :-1], data[:, -1])


def compute_db(data):
    """return davies boulding score"""
    return davies_bouldin_score(data[:, :-1], data[:, -1])


def compute_nmi(labels_true, labels_predict):
    """return adjusted nmi"""
    return adjusted_mutual_info_score(labels_true, labels_predict)


def compute_rand(labels_true, labels_predict):
    """return adjusted rand index"""
    return adjusted_rand_score(labels_true, labels_predict)


def evaluate_all(data, centroids, n_clusters, version, true_labels=None):
    """evaluate clustering evaluation"""
    # Internal Metrics
    error = compute_sse(data, centroids, n_clusters)
    logging.info('Final SSE {}: {}'.format(version, error))
    # print('Final SSE {}: {}'.format(version, error))
    ch_idx = compute_ch(data=data)
    logging.info("Calinski-Harabasz Index {}: {}".format(version, ch_idx))
    # print("Calinski-Harabasz Index {}: {}".format(version, ch_idx))
    db_idx = compute_db(data=data)
    logging.info("Davies Boulding Index {}: {}".format(version, db_idx))
    # print("Davies Boulding Index {}: {}".format(version, db_idx))

    # External Metrics
    if true_labels is not None:
        nmi = compute_nmi(true_labels, tf.cast(data[:, -1], tf.int16))
        logging.info("Normalized Mutual Info Score {}: {}".format(version, nmi))
        # print("Normalized Mutual Info Score {}: {}".format(version, nmi))
        rand = compute_rand(true_labels, tf.cast(data[:, -1], tf.int16))
        logging.info("Rand Index {}: {}".format(version, rand))
        # print("Rand Index v1: {}".format(rand))
    else:
        nmi = None
        rand = None
    return error, ch_idx, db_idx, nmi, rand


def save_all_infos(processor, n_objects, n_features, n_clusters, batch_size_per_replica, centr_fixed, it_fixed,
                   max_iter, iterations, elapsed_time, sse, ch_idx, db_idx, nmi, rand, timestamp):
    """save parameters and clustering evaluation in csv (v2 repeats v1)"""
    header = ["Processor", "Num objects", "Num dimensions", "Num clusters", "Batch size per replica",
              "Initial centroids fixed", "Num iterations fixed", "Max num of iterations", "Num of iterations",
              "Elapsed time", "SSE", "CH index", "DB index", "NMI", "Rand index", "Timestamp"]
    values = [processor, n_objects, n_features, n_clusters, batch_size_per_replica, centr_fixed, it_fixed,
              max_iter, iterations, elapsed_time, sse, ch_idx, db_idx, nmi, rand, timestamp]
    if os.path.exists('evaluation/all_results.csv'):
        with open('evaluation/all_results.csv', 'a', encoding='UTF8', newline='') as f:
            writer = csv.writer(f)
            writer.writerow(values)
    else:
        with open('evaluation/all_results.csv', 'a', encoding='UTF8', newline='') as f:
            writer = csv.writer(f)
            writer.writerow(header)
            writer.writerow(values)
    return


def save_all_infos_sep(processor, version, n_objects, n_features, n_clusters, batch_size_per_replica, centr_fixed,
                       it_fixed, max_iter, iterations, elapsed_time, sse, ch_idx, db_idx, nmi, rand, timestamp):
    """save parameters and clustering evaluation in csv (v2 repeats v1)"""
    header = ["Processor", "Version", "Num objects", "Num dimensions", "Num clusters", "Batch size per replica",
              "Initial centroids fixed", "Num iterations fixed", "Max num of iterations", "Num of iterations",
              "Elapsed time", "SSE", "CH index", "DB index", "NMI", "Rand index", "Timestamp"]
    values = [processor, version, n_objects, n_features, n_clusters, batch_size_per_replica, centr_fixed, it_fixed,
              max_iter, iterations, elapsed_time, sse, ch_idx, db_idx, nmi, rand, timestamp]
    if os.path.exists('evaluation/all_results_sep.csv'):
        with open('evaluation/all_results_sep.csv', 'a', encoding='UTF8', newline='') as f:
            writer = csv.writer(f)
            writer.writerow(values)
    else:
        with open('evaluation/all_results_sep.csv', 'a', encoding='UTF8', newline='') as f:
            writer = csv.writer(f)
            writer.writerow(header)
            writer.writerow(values)
    return


def save_all_infos_comb(processor, n_objects, n_features, n_clusters, batch_size_per_replica, centr_fixed, iter_first,
                        max_iter, iterations, elapsed_time, sse, ch_idx, db_idx, nmi, rand, timestamp, version_first):
    """save parameters and clustering evaluation in csv"""
    header = ["Processor", "Num objects", "Num dimensions", "Num clusters", "Batch size per replica",
              "Initial centroids fixed", "Start version", "Fixed iterations", "Max num of iterations",
              "Actual iterations", "Elapsed time", "SSE", "CH index", "DB index", "NMI", "Rand index", "Timestamp"]
    values = [processor, n_objects, n_features, n_clusters, batch_size_per_replica, centr_fixed, version_first,
              iter_first, max_iter, iterations, elapsed_time, sse, ch_idx, db_idx, nmi, rand, timestamp]
    if os.path.exists('evaluation/all_results_comb.csv'):
        with open('evaluation/all_results_comb.csv', 'a', encoding='UTF8', newline='') as f:
            writer = csv.writer(f)
            writer.writerow(values)
    else:
        with open('evaluation/all_results_comb.csv', 'a', encoding='UTF8', newline='') as f:
            writer = csv.writer(f)
            writer.writerow(header)
            writer.writerow(values)
    return


def save_evaluation_results(init_sse, sse, ch_index, db_index, nmi, rand, timestamp, directory,
                            input_filename, batch_size_per_replica, total_elapsed):
    """save evaluation results and upload to Google cloud bucket"""
    print("Saving evaluation results ...")
    time_formatted = timestamp.strftime("%Y-%m-%d_%H-%M-%S")
    filename = time_formatted + '_' + input_filename + '_evaluation.txt'
    with open(directory + '/' + filename, 'a', encoding='UTF8', newline='') as f:
        f.write('Evaluation results {}\n'.format(timestamp))
        f.write("Parameters:\n")
        f.write("Filename: {}\n".format(input_filename))
        f.write("Batch size per replica: {}\n".format(batch_size_per_replica))
        f.write("Total run time: {}\n\n".format(total_elapsed))
        f.write("Metrics:\n")
        f.write("Initial SSE: {}\n".format(init_sse))
        f.write("SSE: {}\n".format(sse))
        f.write("Calinski-Harabasz Index: {}\n".format(ch_index))
        f.write("Davies Boulding Index: {}\n".format(db_index))
        f.write("Normalized Mutual Index Score: {}\n".format(nmi))
        f.write("Rand Index: {}\n".format(rand))
    print("Evaluation results saved")

    return


def upload_evaluation_results(directory, input_filename, timestamp):
    # upload
    time_formatted = timestamp.strftime("%Y-%m-%d_%H-%M-%S")
    filename = time_formatted + '_' + input_filename + '_evaluation.txt'
    print("Uploading evaluation results ...")
    storage_client = storage.Client()
    buckets = list(storage_client.list_buckets())
    bucket_name = buckets[0].name
    bucket = storage_client.bucket(bucket_name)
    blob = bucket.blob("evaluation/" + filename)
    blob.content_type = "text/csv"
    blob.upload_from_filename(directory + '/' + filename)
    print("Evaluation results uploaded")
