import logging

import tensorflow as tf


class KMeans:
    """K-means computation
        Attributes
        ----------
            n_clusters      -- the number of clusters (k) in the dataset
            n_features      -- number of dimensions for a single data point
            tolerance       -- the dataset
            max_iter        -- maximum iterations

        Methods
        -------
            initialize_centroids    -- return initial centroids
            assign_fn               -- return cluster labels for each data point
            assign_and_prep_fn      -- return new centroids, number of points per cluster, data points with labels
            assign_and_update_fn    -- return final centroids, data points with labels, number of iterations needed
       """
    def __init__(self, n_clusters, n_features, tolerance, max_iter):
        self.n_clusters = n_clusters
        self.n_features = n_features
        self.tolerance = tolerance
        self.max_iter = max_iter

    @tf.function  # comment for main loops
    def initialize_centroids(self, data, centroids_fixed=False) -> tf.slice:
        """ initialize random centroids """
        if centroids_fixed:
            # tf.print("centroids fixed")
            init_centroids = tf.slice(data[:, :-1], [0, 0], [self.n_clusters, -1])
        else:
            shuffled_data = tf.random.shuffle(data)
            init_centroids = tf.slice(shuffled_data[:, :-1], [0, 0], [self.n_clusters, -1])
        logging.info('Centroids initialized; fixed = {}'.format(centroids_fixed))
        return init_centroids

    @tf.function
    def assign_fn(self, points, centroids):
        """assign points to cluster, return assignment"""
        # compute scalar products (m^T) of each centroid with itself
        m = tf.linalg.diag_part(tf.tensordot(centroids, centroids, [[1], [1]]))
        # Column vector of n ones (1_n)
        n = tf.ones(tf.shape(points)[0])
        # calculate 0.5 * 1_n * m^T
        d1 = tf.multiply(tf.constant([0.5]), tf.einsum('i,j->ij', n, m))
        # calculate X*M^T
        d2 = tf.matmul(points, centroids, transpose_b=True)
        # calculate distance D:= 1/2 * 1_n * m^T - X * M^T
        distance = tf.subtract(d1, d2)
        # find minimum distance
        assignments = tf.math.argmin(input=distance, axis=1, output_type=tf.int32)

        # logging.info('Data points assigned')
        return assignments

    @tf.function
    def assign_and_prep_fn(self, data, centroids):
        points = data[:, :-1]
        assignments = self.assign_fn(points, centroids)

        centroids = tf.math.unsorted_segment_sum(points, assignments, self.n_clusters)
        counts = tf.math.unsorted_segment_sum(tf.ones_like(points, dtype=tf.float32), assignments, self.n_clusters)

        assignments_expanded = tf.expand_dims(assignments, 1)
        output = tf.concat([tf.cast(points, tf.float32), tf.cast(assignments_expanded, tf.float32)], 1)

        # logging.info('Data points assigned and centroids prepared')
        # tf.print('Data points assigned and centroids prepared')
        return centroids, counts, output

    @tf.function
    def assign_and_update_fn(self, data: tf.float32, centroids: tf.float32, epochs_fixed) -> tuple:
        points = data[:, :-1]
        assert self.max_iter > 0, "max_iter must be at least 1"
        for it in tf.range(self.max_iter):  # must be tf-loop because of stopping condition
            assignments = self.assign_fn(points, centroids)

            centroids_prev = centroids
            centroids = tf.math.unsorted_segment_mean(points, tf.cast(assignments, tf.int32), self.n_clusters)
            # logging.info('Centroids updated')

            if not epochs_fixed:
                # stopping condition = no change (small tolerance)
                change = tf.square(tf.divide(tf.subtract(centroids, centroids_prev), centroids_prev))
                cx = tf.maximum(change, 1e-9)
                rel_change = tf.sqrt(cx)

                bool_change = tf.math.less_equal(rel_change, self.tolerance)
                num_true = tf.reduce_sum(tf.cast(bool_change, tf.int32))
                # if num_true == self._n_clusters * self._n_features:
                if tf.equal(num_true, tf.multiply(self.n_clusters, self.n_features)):
                    break
        assignments_expanded = tf.expand_dims(assignments, 1)
        output = tf.concat([tf.cast(points, tf.float32), tf.cast(assignments_expanded, tf.float32)], 1)
        return centroids, output, it + 1

    @tf.function
    def predict_fn(self, data, centroids):
        points = data[:, :-1]
        assignments = self.assign_fn(points, centroids)
        assignments_expanded = tf.expand_dims(assignments, 1)
        output = tf.concat([tf.cast(points, tf.float32), tf.cast(assignments_expanded, tf.float32)], 1)
        return output
