import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

import numpy as np

import logging
from datetime import datetime

import tensorflow as tf

import arguments
import authenticate
import data
import evaluation
# from standard_kmeans import KMeans
from kmeans import KMeans

print("Tensorflow version " + tf.__version__)

authenticate.authenticate()

# LOGGING
start_time = datetime.now()
date = datetime.today().strftime("%Y-%m-%d")
timestamp = datetime.today()
timestamp_formatted = timestamp.strftime("%Y-%m-%d %H:%M:%S")
logging.basicConfig(filename='logs/' + date + '_log.log', level=logging.INFO)
logging.info('------------------------ Date and time: {} ---------------------------------'.format(timestamp_formatted))

# VARIABLES
INIT_CENTROIDS_FIXED = False
FILENAME = "data_10240_8_8"
LABELS_EXIST = True
MAX_EPOCHS = 100
TOLERANCE = 0.01  # 0.01, 0.001
BATCH_SIZE_PER_REPLICA = 256

EPOCHS_FIXED = False

input_file_path, tpu_name = arguments.get_terminal_args()

if FILENAME == "bank-additional-full_preprocessed" or FILENAME == "mirai_preprocessed":
    n_clusters = 2
elif FILENAME == "mnist_preprocessed":
    n_clusters = 256
else:
    _, num_objects, num_features, n_clusters = FILENAME.split("_", 3)  # num_objects, num_features
    assert int(num_objects) >= BATCH_SIZE_PER_REPLICA, "Number of objects must be greater or equal than the batch " \
                                                       "size per replica "
    n_clusters = int(n_clusters)

logging.info('Filename: {}'.format(FILENAME))
logging.info('Number of clusters: {}'.format(n_clusters))
logging.info("Tolerance set to: {}".format(TOLERANCE))


def distribute_mean(dataset, centroids):
    """compute k-Means: assignment and update step with distributed training;
        return data points with labels, final centroids"""
    logging.info("Running version 2")

    per_batch_iterations_sum = 0
    per_batch_centroids = 0
    count = 0

    start = datetime.now()
    for x in dataset:
        per_replica_centroids, per_replica_output, per_replica_iterations = strategy.run(clf.assign_and_update_fn,
                                                                                         args=(x, centroids,
                                                                                               EPOCHS_FIXED))
        per_batch_centroids += strategy.reduce(tf.distribute.ReduceOp.SUM, per_replica_centroids, axis=None)
        per_batch_iterations_sum += strategy.reduce(tf.distribute.ReduceOp.SUM, per_replica_iterations, axis=None)

        if strategy.num_replicas_in_sync > 1:
            per_replica_values = tf.convert_to_tensor(per_replica_output.values)
            a, b, c = per_replica_values.shape
            per_batch_output = tf.reshape(per_replica_values, [a * b, c])
        else:
            per_batch_output = tf.convert_to_tensor(per_replica_output)

        if count == 0:
            output = per_batch_output
        else:
            output = tf.concat([output, per_batch_output], axis=0)
        count += 1

    denominator = n_objects // BATCH_SIZE_PER_REPLICA
    centroids_final = per_batch_centroids / denominator

    end = datetime.now()
    elapsed = end - start

    num_iterations_avg = per_batch_iterations_sum / denominator

    logging.info('Elapsed time: {}:'.format(elapsed))
    logging.info("Avg. number of iterations: {} of {} iterations".format(num_iterations_avg, MAX_EPOCHS))
    print("Version 2: ", elapsed)
    print("Avg. number of iterations: {} of {} iterations".format(num_iterations_avg, MAX_EPOCHS))

    return output, centroids_final, num_iterations_avg.numpy(), elapsed


# GET DATASET
data_input, true_labels, n_objects, n_features = data.read_and_expand(filename=FILENAME, path=input_file_path,
                                                                      labels=LABELS_EXIST)

# INITIALIZE CENTROIDS
clf = KMeans(n_clusters, n_features, TOLERANCE, MAX_EPOCHS)
centroids_input = clf.initialize_centroids(data_input, INIT_CENTROIDS_FIXED)

# save initial centroids
filename_initial_centroids = 'init_centroids_' + FILENAME + "_" + timestamp.strftime("%Y-%m-%d_%H-%M-%S")
np.savetxt('results_tmp/' + filename_initial_centroids + '.csv',
           centroids_input, delimiter=',')

init_sse = evaluation.compute_sse(data_input, centroids_input, n_clusters)
logging.info('Initial SSE: {}'.format(init_sse))
print("Initial SSE: ", init_sse)

# FIND AVAILABLE DEVICE, CONNECT AND INITIALIZE STRATEGY
physical_devices_cpu = tf.config.list_physical_devices('CPU')
physical_devices_gpu = tf.config.list_physical_devices('GPU')
try:
    tpu_resolver = tf.distribute.cluster_resolver.TPUClusterResolver(tpu=tpu_name)
except ValueError:
    tpu_resolver = None
if tpu_resolver:
    processor = "TPU"
    tf.config.experimental_connect_to_cluster(tpu_resolver)
    tf.tpu.experimental.initialize_tpu_system(tpu_resolver)
    strategy = tf.distribute.TPUStrategy(tpu_resolver)
else:
    if len(physical_devices_gpu) > 0:
        processor = "GPU"
        print("Num GPUs:", len(physical_devices_gpu))
        strategy = tf.distribute.MirroredStrategy(["GPU:0", "GPU:1", "GPU:2", "GPU:3"])  # 4 replicas
    else:
        processor = "CPU"
        print("Num CPUs:", len(physical_devices_cpu))
        strategy = tf.distribute.get_strategy()  # default distribution strategy in Tensorflow. Works on CPU and
        # single GPU. (1 replica)

logging.info('Running on {}'.format(processor))
print('Running on {}'.format(processor))
logging.info('REPLICAS: {}'.format(strategy.num_replicas_in_sync))
print("REPLICAS: ", strategy.num_replicas_in_sync)

# BATCH DATASET AND DISTRIBUTE
batched_dataset, remainder = data.batch_dataset(dataset=data_input, batch_size_per_replica=BATCH_SIZE_PER_REPLICA,
                                                n_objects=n_objects, num_replicas=strategy.num_replicas_in_sync)
dist_dataset = strategy.experimental_distribute_dataset(batched_dataset)

# REPEAT WITH VERSION 2 (distribute_mean)
version = "distribute_mean (v2)"

final_output, final_centroids, avg_iterations, time = distribute_mean(dist_dataset, centroids_input)

if remainder is not None:
    remainder_output = clf.predict_fn(remainder, final_centroids)
    final_output = tf.concat([final_output, remainder_output], 0)

error, ch_idx, db_idx, nmi, rand = evaluation.evaluate_all(final_output, final_centroids, n_clusters, version,
                                                           true_labels)
evaluation.save_all_infos(processor=processor, n_objects=n_objects, n_features=n_features, n_clusters=n_clusters,
                          batch_size_per_replica=BATCH_SIZE_PER_REPLICA, centr_fixed=INIT_CENTROIDS_FIXED,
                          it_fixed=EPOCHS_FIXED, max_iter=MAX_EPOCHS, iterations=avg_iterations, elapsed_time=time,
                          sse=error, ch_idx=ch_idx, db_idx=db_idx, nmi=nmi, rand=rand, timestamp=timestamp_formatted)

# SHUTDOWN TPU (if exists)
if tpu_resolver:
    tf.tpu.experimental.shutdown_tpu_system(tpu_resolver)

# SAVE RESULTS AND EVALUATE
try:
    final_output, final_centroids
except NameError:
    print("final output and/or final centroids not defined\n Could not finish!")
else:
    print("k-Means finished")
    logging.info("k-Means finished")

    # Save and upload results
    directory = 'results_tmp'
    filename_final_centroids, filename_final = data.save_results(data=final_output, centroids_final=final_centroids,
                                                                 timestamp=timestamp, input_filename=FILENAME,
                                                                 directory=directory)
    data.upload_results(directory=directory, filename_initial_centroids=filename_initial_centroids,
                        filename_final_centroids=filename_final_centroids, filename_final=filename_final)

    timestamp2 = datetime.today().strftime("%Y-%m-%d %H:%M:%S")
    logging.info("Finish time: {}".format(timestamp2))
    end_time = datetime.now()
    total_elapsed = end_time - start_time
    logging.info("Total runtime: {}".format(total_elapsed))
    print("Total runtime: {}".format(total_elapsed))

    evaluation.save_evaluation_results(init_sse=init_sse, sse=error, ch_index=ch_idx, db_index=db_idx, nmi=nmi,
                                       rand=rand, timestamp=timestamp, directory=directory, input_filename=FILENAME,
                                       batch_size_per_replica=BATCH_SIZE_PER_REPLICA, total_elapsed=total_elapsed)
    evaluation.upload_evaluation_results(directory=directory, input_filename=FILENAME, timestamp=timestamp)

print("DONE.")
