import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

import logging
from datetime import datetime

import numpy as np
import tensorflow as tf

import arguments
import authenticate
import data
import evaluation
# from standard_kmeans import KMeans
from kmeans import KMeans

print("Tensorflow version " + tf.__version__)

authenticate.authenticate()

# LOGGING
start_time = datetime.now()
date = datetime.today().strftime("%Y-%m-%d")
timestamp = datetime.today()
timestamp_formatted = timestamp.strftime("%Y-%m-%d %H:%M:%S")
logging.basicConfig(filename='logs/' + date + '_log.log', level=logging.INFO)
logging.info('------------------------ Date and time: {} ---------------------------------'.format(timestamp_formatted))

# VARIABLES
INIT_CENTROIDS_FIXED = False
FILENAME = "data_10240_8_8"
LABELS_EXIST = True
EPOCHS = 5  # 2, 5, 10, 20, 50 (num iterations for first version)
MAX_EPOCHS = 5  # 200 (num iterations for second version)
TOLERANCE = 0.01  # 0.01, 0.001
BATCH_SIZE_PER_REPLICA = 256
VERSION = "both"

epochs_fixed = True
first = 1

if first == 1:
    second = 2
else:
    second = 1
    # switch EPOCHS and MAX_EPOCHS
    interim = MAX_EPOCHS
    MAX_EPOCHS = EPOCHS
    EPOCHS = interim

input_file_path, tpu_name = arguments.get_terminal_args()

if FILENAME == "bank-additional-full_preprocessed" or FILENAME == "mirai_preprocessed":
    n_clusters = 2
elif FILENAME == "mnist_preprocessed":
    n_clusters = 256
else:
    _, num_objects, num_features, n_clusters = FILENAME.split("_", 3)  # num_objects, num_features
    assert int(num_objects) >= BATCH_SIZE_PER_REPLICA, "Number of objects must be greater or equal than the batch " \
                                                       "size per replica "
    n_clusters = int(n_clusters)


logging.info('Filename: {}'.format(FILENAME))
logging.info('Number of clusters: {}'.format(n_clusters))
logging.info("Tolerance set to: {}".format(TOLERANCE))
logging.info("Order: {} -> {}".format(first, second))


def distribute_repeat(dataset, centroids):
    """compute k-Means: assignment step with distributed training, update step on CPU;
    return data points with labels, final centroids"""
    logging.info('Running version 1')

    stop_criterion = False
    start = datetime.now()
    assert EPOCHS > 0, "EPOCHS must be at least 1"
    for ep in range(EPOCHS):
        per_batch_centroids_sum = 0
        per_batch_count = 0
        count = 0
        # iterate over the `tf.distribute.DistributedDataset`
        for x in dataset:
            per_replica_centroids_sum, per_replica_count, per_replica_output = strategy.run(clf.assign_and_prep_fn,
                                                                                            args=(x, centroids))

            per_batch_centroids_sum += strategy.reduce(tf.distribute.ReduceOp.SUM, per_replica_centroids_sum, axis=None)
            per_batch_count += strategy.reduce(tf.distribute.ReduceOp.SUM, per_replica_count, axis=None)

            # concatenate assignments only in last iteration
            if ep + 1 == EPOCHS or stop_criterion:
                if strategy.num_replicas_in_sync > 1:
                    per_replica_values = tf.convert_to_tensor(per_replica_output.values)
                    a, b, c = per_replica_values.shape
                    per_batch_output = tf.reshape(per_replica_values, [a * b, c])
                else:
                    per_batch_output = tf.convert_to_tensor(per_replica_output)

                if count == 0:
                    output = per_batch_output
                else:
                    output = tf.concat([output, per_batch_output], axis=0)

            count += 1

        prev_centroids = centroids
        centroids = per_batch_centroids_sum / per_batch_count

        if not epochs_fixed:
            # stopping condition = no change (small tolerance)
            relative_change = tf.sqrt(
                tf.square(tf.divide(tf.subtract(centroids, prev_centroids), prev_centroids)))
            change_bool = tf.math.less_equal(relative_change, TOLERANCE)
            num_true = tf.reduce_sum(tf.cast(change_bool, tf.int32))
            if num_true == n_clusters * n_features:
                if stop_criterion:
                    break
                else:
                    stop_criterion = True

    end = datetime.now()
    elapsed = end - start
    logging.info('Elapsed time: {}:'.format(elapsed))
    logging.info('Final iteration: no. {} of {}'.format(ep + 1, EPOCHS))
    print("Version 1: ", elapsed)
    print('Final iteration: no. {} of {}'.format(ep + 1, EPOCHS))

    return output, centroids, ep, elapsed


def distribute_mean(dataset, centroids):
    """compute k-Means: assignment and update step with distributed training;
        return data points with labels, final centroids"""
    logging.info("Running version 2")

    per_batch_iterations_sum = 0
    per_batch_centroids = 0
    count = 0

    start2 = datetime.now()
    for x in dataset:
        per_replica_centroids, per_replica_output, per_replica_iterations = strategy.run(clf.assign_and_update_fn,
                                                                                         args=(x, centroids,
                                                                                               epochs_fixed))
        per_batch_centroids += strategy.reduce(tf.distribute.ReduceOp.SUM, per_replica_centroids, axis=None)
        per_batch_iterations_sum += strategy.reduce(tf.distribute.ReduceOp.SUM, per_replica_iterations, axis=None)

        if strategy.num_replicas_in_sync > 1:
            per_replica_values = tf.convert_to_tensor(per_replica_output.values)
            a, b, c = per_replica_values.shape
            per_batch_output = tf.reshape(per_replica_values, [a * b, c])
        else:
            per_batch_output = tf.convert_to_tensor(per_replica_output)

        if count == 0:
            output = per_batch_output
        else:
            output = tf.concat([output, per_batch_output], axis=0)
        count += 1

    denominator = n_objects // BATCH_SIZE_PER_REPLICA
    centroids_final = per_batch_centroids / denominator

    end2 = datetime.now()
    elapsed = end2 - start2

    num_iterations_avg = per_batch_iterations_sum / denominator

    logging.info('Elapsed time: {}:'.format(elapsed))
    logging.info("Avg. number of iterations: {} of {} iterations".format(num_iterations_avg, MAX_EPOCHS))
    print("Version 2: ", elapsed)
    print("Avg. number of iterations: {} of {} iterations".format(num_iterations_avg, MAX_EPOCHS))

    return output, centroids_final, num_iterations_avg.numpy(), elapsed


# GET DATASET
data_input, true_labels, n_objects, n_features = data.read_and_expand(filename=FILENAME, path=input_file_path,
                                                                      labels=LABELS_EXIST)

# INITIALIZE CENTROIDS
clf = KMeans(n_clusters, n_features, TOLERANCE, MAX_EPOCHS)
centroids_input = clf.initialize_centroids(data_input, INIT_CENTROIDS_FIXED)
#np.savetxt('results_tmp/init_centroids_' + timestamp.strftime("%Y-%m-%d_%H-%M-%S") + '.csv', centroids_input, delimiter=',')

init_sse = evaluation.compute_sse(data_input, centroids_input, n_clusters)
logging.info('Initial SSE: {}'.format(init_sse))
print("Initial SSE: ", init_sse)

# FIND AVAILABLE DEVICE, CONNECT AND INITIALIZE STRATEGY
physical_devices_cpu = tf.config.list_physical_devices('CPU')
physical_devices_gpu = tf.config.list_physical_devices('GPU')
try:
    tpu_resolver = tf.distribute.cluster_resolver.TPUClusterResolver(tpu=tpu_name)
except ValueError:
    tpu_resolver = None
if tpu_resolver:
    processor = "TPU"
    tf.config.experimental_connect_to_cluster(tpu_resolver)
    tf.tpu.experimental.initialize_tpu_system(tpu_resolver)
    strategy = tf.distribute.TPUStrategy(tpu_resolver)
else:
    if len(physical_devices_gpu) > 0:
        processor = "GPU"
        print("Num GPUs:", len(physical_devices_gpu))
        strategy = tf.distribute.MirroredStrategy(["GPU:0", "GPU:1", "GPU:2", "GPU:3"])  # 4 replicas
    else:
        processor = "CPU"
        print("Num CPUs:", len(physical_devices_cpu))
        strategy = tf.distribute.get_strategy()  # default distribution strategy in Tensorflow. Works on CPU and
        # single GPU. (1 replica)

logging.info('Running on {}'.format(processor))
print('Running on {}'.format(processor))
logging.info('REPLICAS: {}'.format(strategy.num_replicas_in_sync))
print("REPLICAS: ", strategy.num_replicas_in_sync)

# BATCH DATASET AND DISTRIBUTE
batched_dataset, remainder = data.batch_dataset(dataset=data_input, batch_size_per_replica=BATCH_SIZE_PER_REPLICA,
                                                n_objects=n_objects, num_replicas=strategy.num_replicas_in_sync)
dist_dataset = strategy.experimental_distribute_dataset(batched_dataset)

if first == 1:
    # RUN VERSION 1 (distribute_repeat)
    interim_output, interim_centroids, epoch, interim_time = distribute_repeat(dist_dataset, centroids_input)

    # RUN VERSION 2 (distribute_mean)
    epochs_fixed = False
    final_output, final_centroids, avg_iterations, time = distribute_mean(dist_dataset, interim_centroids)
    num_iterations = avg_iterations

else:
    # RUN VERSION 2 (distribute_mean)
    interim_output, interim_centroids, avg_iterations, interim_time = distribute_mean(dist_dataset,
                                                                                      centroids_input)
    # RUN VERSION 1 (distribute_repeat)
    epochs_fixed = False
    final_output, final_centroids, epoch, time = distribute_repeat(dist_dataset, interim_centroids)
    num_iterations = epoch + 1

if remainder is not None:
    remainder_output = clf.predict_fn(remainder, final_centroids)
    final_output = tf.concat([final_output, remainder_output], 0)

error, ch_idx, db_idx, nmi, rand = evaluation.evaluate_all(final_output, final_centroids, n_clusters, VERSION,
                                                           true_labels) #[:41175]
final_time = interim_time + time
logging.info("Final time k-Means: {}".format(final_time))

evaluation.save_all_infos_comb(processor=processor, n_objects=n_objects, n_features=n_features, n_clusters=n_clusters,
                          batch_size_per_replica=BATCH_SIZE_PER_REPLICA, centr_fixed=INIT_CENTROIDS_FIXED,
                          iter_first=EPOCHS, max_iter=MAX_EPOCHS, iterations=num_iterations, elapsed_time=final_time,
                          sse=error, ch_idx=ch_idx, db_idx=db_idx, nmi=nmi, rand=rand, timestamp=timestamp_formatted,
                          version_first=first)

# SHUTDOWN TPU (if exists)
if tpu_resolver:
    tf.tpu.experimental.shutdown_tpu_system(tpu_resolver)

#np.savetxt('results_tmp/final_centroids_' + timestamp.strftime("%Y-%m-%d_%H-%M-%S") + '.csv', final_centroids, delimiter=',')


print("DONE.")
