import logging

import tensorflow as tf


class KMeans:
    def __init__(self, n_clusters, n_features, tolerance, max_iter):
        self.n_clusters = n_clusters
        self.n_features = n_features
        self.tolerance = tolerance
        self.max_iter = max_iter

    @tf.function  # comment for main loops
    def initialize_centroids(self, data, centroids_fixed=False) -> tf.slice:
        """ initialize random centroids """
        if centroids_fixed:
            # tf.print("centroids fixed")
            init_centroids = tf.slice(data[:, :-1], [0, 0], [self.n_clusters, -1])
        else:
            shuffled_data = tf.random.shuffle(data)
            init_centroids = tf.slice(shuffled_data[:, :-1], [0, 0], [self.n_clusters, -1])
        logging.info('Centroids initialized; fixed = {}'.format(centroids_fixed))
        return init_centroids

    @tf.function
    def assign_fn(self, points, centroids):
        # calculate distance to each centroids (result = array of cluster_num arrays)
        centroids_expanded = tf.expand_dims(centroids, 1)
        d = tf.reduce_sum(tf.square(tf.subtract(points, centroids_expanded)), 2)
        distance = tf.sqrt(tf.maximum(d, 1e-9))
        assignments = tf.math.argmin(input=distance, axis=0)
        # logging.info('Data points assigned')
        return assignments

    @tf.function
    def assign_and_prep_fn(self, data, centroids):
        points = data[:, :-1]
        assignments = self.assign_fn(points, centroids)

        centroids = tf.math.unsorted_segment_sum(points, tf.cast(assignments, tf.int32), self.n_clusters)
        counts = tf.math.unsorted_segment_sum(tf.ones_like(points, dtype=tf.float32), assignments, self.n_clusters)

        assignments_expanded = tf.expand_dims(assignments, 1)
        output = tf.concat([tf.cast(points, tf.float32), tf.cast(assignments_expanded, tf.float32)], 1)
        # logging.info('Data points assigned and centroids prepared')
        return centroids, counts, output

    @tf.function
    def assign_and_update_fn(self, data: tf.float32, centroids: tf.float32, epochs_fixed) -> tuple:
        points = data[:, :-1]
        assert self.max_iter > 0, "max_iter must be at least 1"
        for it in tf.range(self.max_iter):  # must be tf-loop because of stopping condition

            assignments = self.assign_fn(points, centroids)

            centroids_prev = centroids
            centroids = tf.math.unsorted_segment_mean(points, tf.cast(assignments, tf.int32), self.n_clusters)
            logging.info('Centroids updated')

            if not epochs_fixed:
                # stopping condition = no change (small tolerance)
                change = tf.square(tf.divide(tf.subtract(centroids, centroids_prev), centroids_prev))
                cx = tf.maximum(change, 1e-9)
                rel_change = tf.sqrt(cx)

                bool_change = tf.math.less_equal(rel_change, self.tolerance)
                num_true = tf.reduce_sum(tf.cast(bool_change, tf.int32))
                if tf.equal(num_true, tf.multiply(self.n_clusters, self.n_features)):
                    break
        assignments_expanded = tf.expand_dims(assignments, 1)
        output = tf.concat([tf.cast(points, tf.float32), tf.cast(assignments_expanded, tf.float32)], 1)
        return centroids, output, it + 1

    @tf.function
    def predict_fn(self, data, centroids):
        points = data[:, :-1]
        assignments = self.assign_fn(points, centroids)
        assignments_expanded = tf.expand_dims(assignments, 1)
        output = tf.concat([tf.cast(points, tf.float32), tf.cast(assignments_expanded, tf.float32)], 1)
        return output
